/**************************  FIDL.C  *******************************

Author: Rod Crawford.
Date:31/7/89
Mods: 5/ 4/90 Rod Removed Enumerated Types for K&R compatablility
               Now I know how Mitch felt.
               Added #line feature so that syntax
               errors can be traced back to the FIDL.SRC file
               easily.
     29/ 7/90 SFP Added CONSTANT handler.
     26/ 1/91 RC/SFP added %< and >% comment handlers.
      5/ 7/91 JL added multiple input file handling.
              e.g. FIDL fidl.src xcomp.src cgi.src



Desc:

    This program takes a source file of Forth Virtual Machine
    Primitives and generates the required C source code.
    The Forth primitives are defined in FIDL (Forth Internal Definitions
    Language). The syntax of this language is as follows:


    <Forth Name> <Associated C Function name> <Type>
    %
     C Code for Primtive
    %

    N.B. The names and type are delimited by spaces
    e.g.
     ; semicolon I


    Word Types

    The following codes are recognised:

    VC    vocabulary
    CN    constant
    :     colon definition
    :I    immediate colon defintion

    Comment lines may be included in the FIDL source file, these are
    lines which contain \>  in the first two columns. Multiple line comments 
    may be generated using the markers below which must exist between 
    definitions and at the start of the lines.
    %<    start of comment
    >%    end of comment



    The output is placed into three files each with the suffix ".h" for
    generated.
    These files are:
    enum.h      An enumerated type for primitives.
    proto.h     The C Function prototypes.
    extproto.h  The C Function protoypes as external declarations.
    prim.h      The C Function forth machine primitives.
    carve.hThe C Function calls to carve dictionary headers.
    assign.hThe assignments of function pointer to the execution
          vector.


************************************************************************/
#include "cmodel.h"
#include "pinc.h"

#define LINE_LENGTH 255
#define C_NAME_LENGTH  31
#define FORTH_NAME_LENGTH 32
#define TYPE_NAME_LENGTH 10
#define NUMBER_LENGTH 31


/****************************  GLOBAL VAR  ***************************/



FILE *src, *enumerates, *prototypes, *external_prototypes,
     *primitives, *carvers, *assigners;

#if IMPLEMENTATION == RISCOS
FILE *primitives1, *primitives2;
#endif

int line_no=0;
int lineptr=0;
char line[LINE_LENGTH+1];
int enumeration=0;
int file_no;
char *curr_file;


/****************************  ENUMERATIONS  ***************************/


#define IMMEDIATE_COLON 0
#define COLON       1
#define VOC    2
#define CN     3
#define START_COMMENT 4
#define END_COMMENT 5



/*********************FUNCTION PROTOTYPES  ****************************/
#ifdef KNR
int main();
void open_input();
void open_outputs();
void close_outputs();
void close_input();
FILE *must_open();
void must_close();
void process_primitives();
void last_primitive();
void generate_for_colon();
void generate_for_voc();
void generate_for_con();
void process_header();
char *skip_blank_lines_and_comments();
int skip_chars_in_line();
int get_word();
int get_number();
void error();
void generate_enumerate();
void generate_prototype();
void generate_external_prototype();
void generate_colon_carver();
void generate_exe_vec_assignment();
void generate_voc_carver();
void generate_con_carver();
void generate_primitive();
void check_eof();
int type_of();
#else
int main(int argc, char *argv[]);
void open_input( char *input_file );
void open_outputs( void );
void close_input( void );
void close_outputs( void );
FILE *must_open(char *filename,char *mode);
void must_close( FILE *stream );
void process_primitives(void );
void last_primitive( void );
void generate_for_colon(char *c_name,char *forth_name,int category, int *enumeration);
void generate_for_voc(char *forth_name );
void generate_for_con(char *forth_name );
void process_header(char *forth_name,char *c_name,int *category);
char *skip_blank_lines_and_comments(void);
int skip_chars_in_line(char ch);
int get_word(char *word, int max_length );
int get_number(void);
void error(char *message);
void generate_enumerate(char *c_name, int *enumeration);
void generate_prototype(char *c_name);
void generate_external_prototype(char *c_name);
void generate_colon_carver(char *c_name,char *forth_name,int category);
void generate_exe_vec_assignment( char *c_name );
void generate_voc_carver(char *forth_name);
void generate_con_carver(char *forth_name, int number);
void generate_primitive(char *c_name);
void check_eof(char *eof);
int type_of(char *word_type);
#endif





/***************************  MAIN  ***********************************

Desc:Anounces itself, opens the files for IO, processes the FIDL file
     and closes all files. Control may not return to this function if
     the following occurs:
     1) One of the files cannot be opened.
     2) There is a syntactic error in the FIDL file.

***********************************************************************/

main( int argc, char *argv[] )
{
    int i;

    printf( "\nMPE Pinc PowerForth FIDL Compiler V2.20" );
    printf( "\nCopyright (C) 1989, 1990, 1991 MicroProcessor Engineering\n\n" );

    if( argc > 1 ) {
       open_outputs();
       for( i=1; i<argc; i++ ) {
          printf( "Processing: %s\n", argv[i] );
          file_no = i;
          line_no = 0;
          curr_file = argv[i];
          open_input( argv[i] );
          process_primitives();
          close_input();
       }
       last_primitive();
       close_outputs();
       exit( 0 );
    }
    else {
       printf( "Error: No input file(s) specified\n" );
       exit( 1 );
    }
}




/*****************************OPEN_OUTPUTS **************************

Desc:Attempts to open all output files used by this program. Control 
     will not return to this function if one of the files cannot be
     opened.

In:    NONE
Out:   NONE

*********************************************************************/

void open_outputs( void )
{
#if IMPLEMENTATION == RISCOS   
    enumerates =          must_open( "h.enum", "w" );
    prototypes =          must_open( "h.proto", "w" );
    external_prototypes = must_open( "h.extproto", "w" );
    primitives1 =          must_open( "h.prim1", "w" );
    primitives2 =          must_open( "h.prim2", "w" );
    carvers =    must_open( "h.carve", "w" );
    assigners =  must_open( "h.assign", "w" );
#else
    enumerates =          must_open( "enum.h", "w" );
    prototypes =          must_open( "proto.h", "w" );
    external_prototypes = must_open( "extproto.h", "w" );
    primitives =          must_open( "prim.h", "w" );
    carvers =    must_open( "carve.h", "w" );
    assigners =  must_open( "assign.h", "w" );
#endif
}



/*****************************OPEN_INPUT **************************

Desc:Attempts to open an input file used by this program. Control will
     not return to this function if one of the files cannot be opened.

In:    input file name
Out:   NONE

*********************************************************************/

void open_input( char *input_file )
{
   src = must_open( input_file, "r" );
}



/*****************************CLOSE_OUTPUTS **************************

Desc:Attempts to close all output files used by this program. Control 
     will not return to this function if one of the files cannot be
     opened.

In:    NONE
Out:   NONE

*********************************************************************/

void close_outputs( void )
{
#if IMPLEMENTATION == RISCOS
    must_close( enumerates );
    must_Close( prototypes );
    must_close( external_prototypes );
    must_close( primitives1 );
    must_close( primitives2 );
    must_close( carvers );
    must_close( assigners );
#else
    must_close( enumerates );
    must_close( prototypes );
    must_close( external_prototypes );
    must_close( primitives );
    must_close( carvers );
    must_close( assigners );
#endif
}



/*****************************CLOSE_INPUT **************************

Desc:Attempts to close an input file used by this program. Control will
     not return to this function if one of the files cannot be opened.

In:    input file name
Out:   NONE

*********************************************************************/

void close_input( void )
{
   must_close( src );
}



/*****************************MUST_OPEN  ***************************

Desc:Attempts to open the file <filename> in the required mode.
     If it succeed the stream pointer is returned.
     If the file cannot be opened, it is deemed impossible for
     further execution of the program to continue and execition is
     terminated with the appropriate error message.

In:  filename    pointer to string containing filename
     mode     pointer to string defining mode in which the file
              should be opened.

Out: stream        pointer to file stream opened.

**********************************************************************/

FILE *must_open( filename, mode )
char *filename, *mode;
{
    FILE *stream;

    if (( stream = fopen( filename, mode )) == NULL )
    {
       fprintf( stderr, "\nFile %s\n", filename );
       error( "Can't Open File" );
    }
    return( stream );
}



/*****************************MUST_CLOSE  ***************************

Desc:Attempts to close the file <filename>.
     If the file cannot be closed, it is deemed impossible for
     further execution of the program to continue and execution is
     terminated with the appropriate error message.

In:  filename    pointer to string containing filename
Out: NONE

**********************************************************************/

void must_close( stream )
FILE *stream;
{
    if (( fclose( stream )) == EOF )
    {
       error( "Can't close File" );
    }
}



/***************************  PROCESS_PRIMITIVES  **********************

Desc:   Whilst there is input the FIDL file is processed:
     The header is for each primitive is scanned and the
     appropriate entry in each of the generated files is
     constructed depending upon the category of primitive specified.
     Execution may not return to this function if the following
     holds true:

     1)  There is a syntax error in the header of the primitive
         entry.
     2)  There is a syntax error in the body of the primitive
         entry.

In:    NONE
Out:   NONE

************************************************************************/

void process_primitives( void )
{
    char c_name[ C_NAME_LENGTH ],
      forth_name[ FORTH_NAME_LENGTH ];
    int category;

    while( skip_blank_lines_and_comments() ) {
     process_header( forth_name, c_name, &category );
     switch( category ) {
       case IMMEDIATE_COLON : generate_for_colon( c_name, forth_name,
                                   category, &enumeration );
                     break;
       case COLON          : generate_for_colon( c_name, forth_name,
                                   category, &enumeration );
                     break;
       case VOC       : generate_for_voc( forth_name );
                     break;
       case CN        : generate_for_con( forth_name );
                     break;
     }
    }
}



/********************************************************************

Desc: Generate last primitive. This primitive MUST be the last one
      genertated.

In:   NONE
Out:  NONE

*********************************************************************/

void last_primitive()
{
    generate_enumerate( "null_primitive", &enumeration );
}



/*********************GENERATE_FOR_COLON  **************************

Desc: Generates the appropriate entries for a colon definition

In:  c_name  name of the forth primitive in C terms
     forth_name   name of the forth primitive in forth terms
     category  type of the forth primitive.
     enumeration  enumeration variable which will be incremented

Out: NONE

*********************************************************************/

void generate_for_colon(  c_name, forth_name, category, enumeration )
char c_name[ C_NAME_LENGTH ],
     forth_name[ FORTH_NAME_LENGTH ];
int category, *enumeration;
{
     generate_enumerate( c_name, enumeration );
     generate_prototype( c_name );
     generate_external_prototype( c_name );
     generate_colon_carver( c_name, forth_name, category );
     generate_exe_vec_assignment( c_name );
     generate_primitive( c_name );
}





/*********************GENERATE_FOR_VOC  **************************

Desc: Generates the appropriate entries for a colon definition

In:  c_name name of the forth primitive in C terms
     forth_name  name of the forth primitive in forth terms
     category type of the forth primitive.

Out: NONE

*********************************************************************/

void generate_for_voc(forth_name )
char forth_name[ FORTH_NAME_LENGTH ];
{
     generate_voc_carver( forth_name );
}





/* *******************GENERATE_FOR_CON  ***********************

Desc: Generates the appropriate entries for a colon definition

In:  forth_name  name of the forth primitive in forth terms

Out: NONE

******************************************************************* */

void generate_for_con( forth_name )
char forth_name[ FORTH_NAME_LENGTH ];
{
     int number;

     number = get_number();
     generate_con_carver( forth_name, number );
}




/******************************  PROCESS_HEADER  **************************

Desc:Given a line of input which "should" contain a header the following
     strings are formed:

        1) The name of the Forth primitive in Forth terms e.g. >R
        2) The name of the forth primive in C terms e.g. to_r
        3) The type of the forth primitive e.g. colon defintion.

     Exection may not return to this function if the input line
     is exhausted prior to the construction of any of the strings.

In:  NONE

Out: forth_name  name of the forth primive in forth terms
     c_name        name of the forth primitive in C terms
     category    type of the forth primitive.

******************************************************************/


void process_header( forth_name, c_name, category )
char c_name[ C_NAME_LENGTH ],
     forth_name[ FORTH_NAME_LENGTH ];
int *category;

{
    char word_type[ TYPE_NAME_LENGTH ];

    get_word( forth_name, FORTH_NAME_LENGTH );
    get_word( c_name, C_NAME_LENGTH );
    get_word( word_type, TYPE_NAME_LENGTH );
    *category = type_of( word_type );
}





/********************  SKIP_BLANK_LINES_AND_COMMENTS 
***********************

Desc:Reads the source file skipping blank lines. During this
     process the stream may become exhausted or an I/O error
     may occur. This value is returned by the function.

In:  NONE
Out: eof  file stream status.

************************************************************************/

char* skip_blank_lines_and_comments()
{
    char *eof;
    int skip_line, in_comment;

    skip_line = 1;
    lineptr = 0;
    in_comment = 0;

    while(skip_line ) 
    {
      line_no++;
      eof = fgets( line, LINE_LENGTH, src );

      if(!eof) { return( eof ); }

      if( in_comment ) 
      {
          if((line[0] == '>') && (line[1] == '\%' ))
                in_comment = 0;
      }
      else 
      {   
           if( line[0] == '\n' )
           { skip_line = 1; }
           else
                if( (line[0] == '\\') &&
                   !(( line[1] == '\\' ) || ( line[1] == '\"' )))
                { skip_line = 1;}
                else
                {
                    if((line[0] == '\%') && (line[1] == '<' ))
                    {
                         in_comment = 1;
                         skip_line = 1;
                    }
                    else 
                    {     skip_line = 0;}
                }
      }
    }
    return( eof );
}




/*    do {
       line_no++;
    }
    while(( eof = fgets( line, LINE_LENGTH, src )) &&
      (( line[0] == '\n' ) ||
       (( line[0] == '\\' ) &&
         !(( line[1] == '\\' ) || ( line[1] == '\"' )) )));
    lineptr = 0;
    return( eof );
}
*/




/***************************  SKIP_CHARS_IN_LINE  **********************

Desc:given current line pointer poining into a given line- the
     line pointer is advanced whilst line[ lineptr ] == ch and
     we have not reached the end of the line.

In:  ch   character to be skipped over.

Out: lineptr the new value of the line pointer is returned as
          the functions return value.
          this is:  n where n == srlen( line ) OR
                 n where NOT( line[ n ] == ch )

**********************************************************************/

int skip_chars_in_line( ch )
char ch;
{
    while( lineptr < strlen( line ) && line[ lineptr++ ] == ch );
    return( --lineptr );
 }




/**************************  GET_WORD  *********************************

Desc:given a line of input and a pointer into it, this function
     attempts to get the next word (delimited by spaces)
     from the line and place it in word. The size of the word
     max not exceed the max_length string provided. If it does
     an error occurs. Control will not be returned by this function
     if the following occurs:

     1) No word can be found in the input line before the
        line is exhausted.
     2) The word in the input line is longer than the permitted
        length.

     The new value of the line pointer is returned. This points
     to the character after the line last character of the word.

In:  max_length  max length of the word that can be constructed.
              including the terminating \0.

Out: word     the word string formed terminated by \0.
     lineptr     the new vlaue of the lineptr is reuturned by the
              function.

**********************************************************************/

int get_word( word, max_length )
char *word;
int max_length;
{
    int i=0;

    skip_chars_in_line( ' ' );

    if( lineptr == strlen( line ))
     error( "unexpected end of line");

    while( (i < max_length) && (line[ lineptr ] != ' ' )
                   && (line[ lineptr ] != '\n' ))
     word[ i++ ] = line[ lineptr++ ];

    if( i == max_length )
     error( "ID too long" );

    word[ i ] = '\0';

    return( lineptr );
}



/* **************************** GET_NUMBER ******************************

Desc:Gathers a number from the current input line, and returns it as
     an INT. An error is reported if conversion fails for any reason.

In:  NONE

Out: number    the collected INT.

********************************************************************** */

int get_number()
{
     char number_buff[ NUMBER_LENGTH ];
     int number;

     get_word( number_buff, NUMBER_LENGTH );
     if (sscanf( number_buff, "%d", &number ) != 1)
        error("Valid number needed");
     return( number );
}


/******************************  ERROR*********************************

Desc:prints the given given line and the specified error
     message pointing to the related part of the line.
     Execution is then terminated with an exit code of 1.


In:   message the error message

Out: NONE

*************************************************************************/

void error( message )
char message[];
{
    int i;
    
    fprintf( stderr, "\nLine # %d", line_no );
    fprintf( stderr, "\n%s", line );
    for( i=0; i<lineptr-1; i++ )  fprintf( stderr, "-" );
    fprintf( stderr, "^ %s \n\n", message );
    exit( 1 );
}





/***********************  GENERATE_ENUMERATE  *************************

Desc: generates an entry in the enumerates file of the form:
         #define prim_<C_NAME> n

    e.g. given a C name of forth the forth primitive >R of to_r
      the following would be generated:
         #define prim_to_r 42

In:  c_name         name of the forth primitive as a c function.
     enumeration  address of the enumeration variable used to
               generate the #define

Out: NONE

*************************************************************************/

void generate_enumerate( c_name, enumeration )
char *c_name;
int *enumeration;
{
    fprintf( enumerates, "#define prim_%s %d\n", c_name, (*enumeration)++ );
}





/***********************  GENERATE_PROTOTYPE  *************************

Desc: generates an entry in the prototypes file of the form:
         void <C_NAME>( void );

    e.g. given a C name of forth the forth primitive >R of to_r
      the following would be generated:
         void to_r( void );

In:  c_name    name of the forth primitive as a c function.

Out: NONE

*************************************************************************/

void generate_prototype( c_name )
char *c_name;

{
    fprintf( prototypes, "  void %s( void );\n", c_name );
}





/*********************GENERATE_EXTERNAL_PROTOTYPE 
*********************

Desc: generates an entry in the external prototypes file of the form:
         extern void <C_NAME>( void );

    e.g. given a C name of forth the forth primitive >R of to_r
      the following would be generated:
         extrern void to_r( void );

In:  c_name    name of the forth primitive as a c function.

Out: NONE

*************************************************************************/

void generate_external_prototype( c_name )
char *c_name;
{
    fprintf( external_prototypes, "  extern void %s( void );\n", c_name );
}






/***********************  GENERATE_COLON_CARVER 
*****************************

Desc: generates an entry in the carvers file of the form:
         compile_colon( "<FORTH_NAME>", (CELL) prim_<C_NAME> );

    e.g. given a C name of the forth primitive >R of to_r
      the following would be generated:
         compile_colon( ">R", (CELL) prim_to_r );

    If the primitive is immediate this is followed by the function call
    immediate();

    e.g. given the C name of the forth primitive IF of if_statemtent
      and that this primitive is immediate, the following would be
      generated:
         compile_colon( "IF", (CELL) prim_if_statement );
         immediate();


In:  c_name        name of the forth primitive as a c function.
     forth_name  name of the forth primitive in forth terms.
     category    type of the forth primitive.

Out: NONE

*************************************************************************/

void generate_colon_carver( c_name, forth_name, category )
char *c_name, *forth_name;
int category;
{
    fprintf( carvers, "  compile_colon( \"%s\", (CELL) prim_%s );\n",
                forth_name, c_name );

    if( category == IMMEDIATE_COLON )
     fprintf( carvers, "  immediate();\n" );


}





/***********************  GENERATE_VOC_CARVER 
*****************************

Desc: generates an entry in the carvers file of the form:
         compile_voc( "<FORTH_NAME>" );

    e.g. given a the forth vocabulary name root
      the following would be generated:
         compile_voc( "root" );

In:  forth_name  name of the forth primitive in forth terms.

Out: NONE

*************************************************************************/

void generate_voc_carver( forth_name )
char *forth_name;
{
    fprintf( carvers, "  compile_voc( \"%s\" );\n",
                forth_name );

}





/***********************  GENERATE_CON_CARVER 
*****************************

Desc: generates an entry in the carvers file of the form:
         compile_con( "<FORTH_NAME>", literal );

    e.g. given the Forth constant name C/L
      the following would be generated:
         compile_con( "C/L", 64 );

In:  forth_name  name of the forth primitive in forth terms.
     number        value of the constant

Out: NONE

*************************************************************************/

void generate_con_carver( forth_name, number )
char *forth_name;
int number;
{
    fprintf( carvers, "  compile_con( \"%s\", %d );\n", forth_name, number );
}



/***********************  GENERATE_EXEVEC_ASSIGNMENT 
*****************************

Desc: generates an entry in the assigners file of the form:
         execution_vector[ prim_<C_NAME> ] = (* C_NAME);

    e.g. given a C name of the forth primitive >R of to_r
      the following would be generated:
         execution_vector[ prim_to_r ] = (* to_r);


In:  c_name        name of the forth primitive as a c function.

Out: NONE

*************************************************************************/

void generate_exe_vec_assignment( c_name )
char *c_name;
{
    fprintf( assigners, "  execution_vector[ (int) prim_%s ] = %s;\n",
               c_name, c_name );
}







/***********************  GENERATE_PRIMITIVE  *************************

Desc: generates an entry in the primitives file of the form:
         void <C_NAME>( void )
         {
          C CODE
         }

    e.g. given a C name of forth the forth primitive >R of to_r
      the following would be generated:
         void to_r( void )
         {
          Associated C Code
         }
    The line number and filename of the fidl file is also output
    so that the C compiler will effectively be in step with the
    fidl file when it is compiling the primitives.

    This function will not return if there is a syntactic error
    in the specification of the C code following the Forth
    primitive header specification. e.g. the next line after the
    header specification does  has not got a % as the first
    character or a premature EOF or I/O error is detected.
    The syntax and semantics of the actual C code for the primitive
    are not checked by this function.

In:  c_name    name of the forth primitive as a c function.

Out: NONE

*************************************************************************/

void generate_primitive( c_name )
char *c_name;
{
    char *eof;

#if IMPLEMENTATION == RISCOS
    static int toggle=1;

    if( toggle ) primitives = primitives1;
    else         primitives = primitives2;
    toggle = !toggle;
#endif

    fprintf( primitives, "#line %d \"%s\"\n", line_no, curr_file );
    fprintf( primitives, "void %s( void )\n{\n", c_name );

    eof = fgets( line, LINE_LENGTH, src ); line_no++;
    check_eof( eof );

    if( line[ 0 ] != '%' )
     error ( "C Source Code Expected" );
    while(( eof = fgets( line, LINE_LENGTH, src )) &&
      ( line[0] != '%' )) {
       line_no++;
       fprintf( primitives, "%s", line );
    }
    line_no++;
    check_eof( eof );

    fprintf( primitives, "}\n\n" );
}




/******************************  CHECK_EOF  *******************************

Desc: Checks to see if an end of file has occurred. If it has,
      the program is terminated.

In:  eof file stream status.

Out: NONE

*************************************************************************/

void check_eof( eof )
char *eof;
{
    if ( eof == NULL )
       error( "Unexpected End Of File" );
}




/******************************  TYPE_OF  ******************************

Desc: Returns type of word specified by the string word_type.
      If no valid word type is specified by the string, an
      appropriate error message is given and execution is terminated.

In:  word_type string defining the type of word.
     line line in which word type was defined (used by
          error handler)

     lineptr pointer into line where word type string ended.

Out: type_of enumerated type of the word.

***********************************************************************/

int type_of( word_type )
char *word_type;
{
     char ch0, ch1;

     ch0 = word_type[ 0 ];
     ch1 = word_type[ 1 ];

     if ((ch0==':') && (ch1!='I'))
        return COLON;
     if ((ch0==':') && (ch1=='I'))
        return IMMEDIATE_COLON;
     if ((ch0=='C') && (ch1=='N'))
        return CN;
     if ((ch0=='V') && (ch1=='C'))
        return VOC;

     error( "Illegal Word Type" );
}

