/*************************************************************************

        Pinc Main Program

Author: Rod Crawford

Date: 16/11/89
Mods:  5/ 4/90 Rod Replaced enum with #defs for K&R compatability
             I now sympathise with Mitch very much.
      12/ 4/90 Rod Added support for OS9 Note the ordering of
             globals is critical for OS9. The big arrays
             have to come at the end.

Definition of the Forth Virtual Machine

**********************************************************************/

#include "cmodel.h"
#include "pinc.h"
#include "lowram.h"
#include "vmsize.h"
#include "fname.h"
#include "misc.h"
#include "strsizes.h"
#include <stdlib.h>

/*************************** Data Types *********************************

An instance of the the primitive enumerated type can hold any of the
Pinc primitive instructions. The enumerations that define the instruction
set of the virtual machine are generated by the FIDL compiler and placed
in the file enum.h

The typedef function_pointer declares an analogue for the rather unwieldy
pointer to function type. A pointer to a function declared using
function_pointer may not return a value, nor may it take a value.

*/


#include "enum.h"

typedef void (*function_pointer) (void);



/******************* External Function Prototypes  ************************/

#ifdef KNR
extern void init_io();
extern int load_dictionary();
extern void compile_primitives();
#ifdef CWEED
extern void init_user_variables();
extern void cold_start();
#endif
extern void inner_interpreter();

#ifdef CHECK_STACKS
extern unsigned CELL pop_data();
#else
#include "stacks.h"
#endif

extern void assign_the_primitives();
extern struct from_fileblock make_from_fileblock();

#else	/* Not KNR */

extern void init_io(void );
extern int load_dictionary(char *filename);
extern void compile_primitives(void );
#ifdef CWEED
extern void init_user_variables(void );
extern void cold_start(void );
#endif
extern void inner_interpreter(void );

#ifdef CHECK_STACKS
extern unsigned CELL pop_data(void );
#else
#include "stacks.h"
#endif

extern void assign_the_primitives(void );
extern struct from_fileblock make_from_fileblock( char* filename, FILE *stream );

#endif    /* KNR */

#include "extproto.h"




/***********************  Function Prototypes  *****************************/

#ifdef KNR
int main();
int bootstrap();
#ifdef CWEED
void make_dictionary();
#else
void run_pinc();
#endif
void clean_ram();
#else
int main(int argc, char *argv[] );
int bootstrap( char *dictionary_filename );
#ifdef CWEED
void make_dictionary( void );
#else
void run_pinc( void );
#endif
void clean_ram(void );
#endif

/*******************  Forth Primitive Function Prototypes ******************

These functions define function prototypes of the primitive operations
that the Pinc virtual machine can execute.
They are generated by the FIDL compiler and placed in the file proto.h.

*/

/* #include "proto.h" */




/*************************** Global Data ********************************

The following defines all global data defintions in the Pinc virtual
machine. Since the majority of fucntions access this data it was felt that
there we little benefit in making them local to main and passign them
as parameters.
*/


/*

Registers:
The Pinc Virtual Machine has the following Registers:

sp        Data Stack Pointer.
rp        Return Stack Pointer.
dp        Dictionary Pointer.
ip        Instruction Pointer.
latest    Pointer to the nfa of the last word defined
voc_link  Pointer to the cfa of the previously defined vocabulary.

*/


unsigned CELL sp, rp, dp, ip, latest, voc_link;



/*

Flags:

The Pinc Virtual Machine Uses The Following Flags:

compiling     True if compiling else False.
kill_switch   Used by Cweed Forth to allow the inner interpreter to
              terminate if we are returning to the top level interpret
              loop written in C.


*/
unsigned CELL compiling;

#ifdef CWEED
int kill_switch;
#endif


/*

I/O Streams

The two from_fileblocks input and output are used throughout Pinc, thus
allowing redirection away from stin and stdout. The from_fileblocks are
declared in io.c and are imported here.

*/

extern struct from_fileblock {
     char filename[ FROM_FILENAME_LENGTH ];
     int page_number, line_number, line_in_page_number,
         last_page_required;
     FILE *stream;
       };

extern struct from_fileblock input, output;


/*

Misc:

current_sp Used to hold the current value of the the data stack pointer
           during a compilation. The current value of sp is stored in
           this variable by the colon compiler. The value is then checked
           by semicolon to see if it concurs with the value of sp at the
           its time of execution. If it does not then there may be an error
        in the current words definition e.g. an IF without an ENDIF.

environment is used to hold the current stack environment for restoration
         after a CELL jump.

*/

unsigned CELL current_sp;
jmp_buf environment;

/*
Memory Areas:

The Pinc Virtual Machine Consists of 3 Memory Areas. These are:
Data Stack, Return Stack and Ram. The user area is defined as the lower
portion of the Ram Area.


Execution Vector.

The execution vector is used to hold the functions which comprise the
Pinc/CWeed Forth primitives. The size of this array is determined by the
maximum number of primitives declared in the fidl.src file.
However since the enumerated type (implemented by #defs) has prim_null_primtive
as its last enumeration, we use this to declare the maximum size of the
array and it will always be correct.

*/
unsigned CELL stack_underflow_buffer1[ 10 ];
unsigned CELL data_stack[ DATA_STACK_TOP ];
unsigned CELL stack_underflow_buffer2[ 10 ];
unsigned CELL return_stack[ RETURN_STACK_TOP ];
unsigned CELL stack_overflow_buffer3[ 10 ];
#if IMPLEMENTATION == MicrosoftC
    unsigned char huge ram[ RAM_TOP ];
#else
#if IMPLEMENTATION == OS9
    remote unsigned char ram[ RAM_TOP ];
#else
    unsigned char ram[ RAM_TOP ];
#endif
#endif



function_pointer execution_vector[ prim_null_primitive ];



/*******************************  main  ****************************

Desc: This function boostraps Pinc.


********************************************************************/

// main(int argc, char *argv) {
main(int argc, char **argv) {
    char *lexicon=(char *)NULL;

/*
    printf("===== main\n");
    printf("\nGeneral data types - MinGW64\n");
    printf(" long long size = %d\n",(int)sizeof(long long));
    printf(" long size      = %d\n",(int)sizeof(long));
    printf(" int size       = %d\n",(int)sizeof(int));
    printf(" short size     = %d\n",(int)sizeof(short));
    printf(" char size      = %d\n",(int)sizeof(char));
    printf(" CELL size      = %d\n", sizeof(CELL));
    printf(" CELL_SIZE      = %d\n", CELL_SIZE);
    printf(" COLD_VECTOR    = %d\n", COLD_VECTOR);
    printf(" WARM_VECTOR    = %d\n", WARM_VECTOR);
    printf("ip              = 0x%llx\n", (long long) ip);
    printf("&RAM            = 0x%llx\n", (long long) &ram );
    printf("prim_null_primitive = 0x%llx\n", prim_null_primitive );
    printf("argc %d\n", argc);
#ifdef CWEED
    printf("CWEED defined\n");
#endif
#ifdef CHECK_INNER
    printf("CHECK_INNER defined\n");
#endif
*/

    if( argc > 1 ) {
        exit( bootstrap( argv[1] ));
    } else {
        lexicon = (char *)getenv( "LEXICON" );

        if( lexicon == (char *)NULL ) {
            exit( bootstrap( DEFAULT_DICTIONARY_FILE ));
        } else {
            exit( bootstrap( lexicon ));
        }
    }
}


/*****************************  bootstrap  *************************

Desc: Either loads up a dictionary  and kicks of Pinc, or
      if one can't be found kicks in CWeed Forth to make one.
      CALL THIS FUNCTION FROM AN APPLICATION IF YOU WISH TO
      START PINC FROM SCRATCH FROM WITHIN THE APPLICATION.

Alg: The environment is saved to allow CELL jumps to return here.
     e.g. when the user types bye.
     the retrun status returned by bootstrap is one less than that
     returned by the setjmp/longjmp combination. This is to accomodate
     the fact that longjmp cannot return a status of zero, however
     a code of zero is required to indicate normal program termination.
     IO is initialised to come from standard input and go to standard
     output.
     The Pinc/Weed Forth primtives are assigned into the execution
     vector.
     If no dictionary file can be loaded:
        we make one using cweed forth
     else
        using the loaded dictioanry, run pinc.

     N.B. As is customary with MPE Forth systems COLD does not
     initialise all the user variables and it may be that if some
     are set during normal forth execution and then COLD executed
     the values they had prior to COLD will remain.

********************************************************************/

int bootstrap( dictionary_filename )
char *dictionary_filename;
{
    int return_status;
    if( (return_status = setjmp( environment )) != 0)
      return( return_status-1 );
    init_io();

#ifdef DONGLE
    test_dongle();
#endif

#if IMPLEMENTATION == SUN
/*  ag-special. The following is commented out for compatibilty
               with SunOS 3.x. Can be uncommented for SUNOS 4.0
               to provide proper key? etc.
*/
    init_signals();
#endif
    assign_the_primitives();

#ifdef CWEED
    make_dictionary();
#else
    if( load_dictionary( dictionary_filename )) {
      run_pinc();
    } else {
      fprintf( stderr, "Can't Find Dictionary File: %s\n",
                dictionary_filename );
      longjmp( environment, 253 );
    }
#endif
}


#ifdef CWEED
/************************  make_dictionary  *****************************

Desc: Starts Cweed Forth loading the CSCAPE file. The CSCAPE file
      contains commands to save the dictionary and exit. If the
      cscape file cannot be found the program exits.

      To start CWeed Forth DP is set above the user variable space,
      the voc_link and latest are zeroed and the primitves are compiled
      into the dictionary. The uservariables (such as S0 etc. are then
      set up. Finally with cscapes file stream on top of the from stack
      cold start is called, making input start from cscape.

************************************************************************/

void make_dictionary( void )
{
    FILE *cscape_file;

    dp = DICTIONARY_START;
    voc_link = 0;
    latest = 0;
    compile_primitives();
    init_user_variables();
    if(( cscape_file = fopen( SECONDARIES_FILE, "r" )) == NULL ) {
        fprintf( stderr,
"Can't Find Cweeds Secondary Code And Programming Environment file CSCAPE\n" );
     longjmp( environment, 255 );
    }
    else {
#if IMPLEMENTATION == OS9
         push_from_fileblock( input );
         input.stream = cscape_file;
         strcpy( input.filename, SECONDARIES_FILE );
         input.line_number = 0;
         input.page_number = 0;
         input.line_in_page_number = 0;

#else
     if( setvbuf( cscape_file, NULL, _IOFBF, 10240 )) {
         fprintf( stderr,
               "Can't Allocate Input Buffer for CSCAPE\n" );
         longjmp( environment, 255 );
      }
     else {
         push_from_fileblock( input );
         input = make_from_fileblock( SECONDARIES_FILE, cscape_file );
     }
#endif
    }
    cold_start();
}
#endif

#ifndef CWEED
/***************************  run_pinc  ********************************

Desc: This function runs full pinc assuming a dictionary has been
      loaded into memory and the COLD_VECTOR has been set to point at
      somewhere sensible as a result of this.

      The stacks are set up to be empty.
      The contents of the cold vector is placed into the instaction
      pointer

        AND AWAY WE GO!!


************************************************************************/

void run_pinc( void )
{
        sp = (CELL) -1;
        rp = (CELL) -1;
        push_data( (CELL) COLD_VECTOR );
        fetch();
        ip = pop_data();
        inner_interpreter();
}
#endif


/*******************************  clean_ram  ***************************

Desc: Fills all ram with Zero

***********************************************************************/

void clean_ram( void )
{
    unsigned CELL i;

    for( i=0; i<RAM_TOP; i++ )ram[ i ] = 0;
}


